/*
 *    author: Lloyd Dupont
 *    mail: lloyd@galador.net
 *    Copyright (C) 2000 Lloyd Dupont
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * this class provide a way to post for later a unique action. as the action
 * is unique, if you post an other action before current one is executed, the
 * current one will be discarded.
 *
 * @see #dispose
 * @see #test
 */
public class DelayedAction
{
    Thread threadTest;
    long timeTest = -1;
    Action action;

    DelayedAction() {}

    protected void init() {
        if(threadTest != null)
            dispose();
        threadTest = null;

        threadTest = new Thread(new Loop());
        threadTest.start();
    }
    boolean running() { return threadTest != null; }

	/**
	 * "terminate" the DelayedAction object. As it is always referenced by
	 * an internal thread this method should be called to free memory.
	 */
    public void dispose() {
        timeTest = -1;
        if(threadTest != null)
            threadTest.interrupt();
        threadTest = null;
    }
	/**
	 * "test nothing", this remove current action.
	 * it is the same than <B>test</B>(<I>null</I>).
	 */
    public void test() { test(null, 600); }
	/** simply call <B>test</B>(<I>a</I>, 600) */
    public void test(Action a) { test(a, 600); }
	/**
	 * post current test. older test will be removing without being executed.
	 *
	 * @param a the action to be executed. if null remove any test
	 * @param delay the time to wait (in millisecond) before doing test.
	 */
    public void test(Action a, long delay) {
        if(a==null && timeTest < 0) {
			action = null;
            return;
		}
        if(threadTest == null)
            init();
		if(a != null && a.equals(action))
			return;

        action = a;
		if(action != null)
	        timeTest = System.currentTimeMillis() + delay;
        else
            timeTest = -1;
        synchronized (threadTest) { threadTest.notifyAll(); }
    }

    class Loop implements Runnable {
        public void run() {
            Thread current = Thread.currentThread();
            while(!Thread.interrupted()) {
                if(timeTest < 0 || action == null)
                    synchronized (current) {
                        try { if(timeTest < 0) current.wait(); }
                        catch(InterruptedException ie) {}
                    }
                else {
                    long l = System.currentTimeMillis();
                    if(l >= timeTest) {
                        if(action != null)
                            action.action(null);
                        action = null;
                        timeTest = -1;
                    }
                    else {
                        synchronized (current) {
                            try { current.wait(timeTest - l); }
                            catch(InterruptedException ie) {}
                        }
                    }
                }
            }
        }
    }
}
